"""
Patches :class:`~intents.model.entity.Sys.Person`

Sources:

* https://www.ssa.gov/oact/babynames/ (top 100)
* https://www.istat.it/it/dati-analisi-e-prodotti/contenuti-interattivi/contanomi
  (top 50)
* https://www.behindthename.com/top/lists/netherlands/2011
* https://www.beliebte-vornamen.de/jahrgang/j2020/top-500-2020
* https://www.ine.es/dyngs/INEbase/es/operacion.htm?c=Estadistica_C&cid=1254736177009&menu=resultados&idp=1254734710990


"""
from typing import List, Union

from intents import Entity, LanguageCode
from intents.language.entity_language import make_language_data

# TODO: limited to 100 names per country. Stats may be about baby names rather
# than whole population. Chinese should be reviewed and expanded.

class I_IntentsFirstName(Entity):
    __intents_internal__ = True
    __entity_language_data__ = {
        LanguageCode.ENGLISH: make_language_data(["Liam", "Noah", "Oliver", "Elijah", "William", "James", "Benjamin", "Lucas", "Henry", "Alexander", "Mason", "Michael", "Ethan", "Daniel", "Jacob", "Logan", "Jackson", "Levi", "Sebastian", "Mateo", "Jack", "Owen", "Theodore", "Aiden", "Samuel", "Joseph", "John", "David", "Wyatt", "Matthew", "Luke", "Asher", "Carter", "Julian", "Grayson", "Leo", "Jayden", "Gabriel", "Isaac", "Lincoln", "Anthony", "Hudson", "Dylan", "Ezra", "Thomas", "Charles", "Christopher", "Jaxon", "Maverick", "Josiah", "Isaiah", "Andrew", "Elias", "Joshua", "Nathan", "Caleb", "Ryan", "Adrian", "Miles", "Eli", "Nolan", "Christian", "Aaron", "Cameron", "Ezekiel", "Colton", "Luca", "Landon", "Hunter", "Jonathan", "Santiago", "Axel", "Easton", "Cooper", "Jeremiah", "Angel", "Roman", "Connor", "Jameson", "Robert", "Greyson", "Jordan", "Ian", "Carson", "Jaxson", "Leonardo", "Nicholas", "Dominic", "Austin", "Everett", "Brooks", "Xavier", "Kai", "Jose", "Parker", "Adam", "Jace", "Wesley", "Kayden", "Silas", "Olivia", "Emma", "Ava", "Charlotte", "Sophia", "Amelia", "Isabella", "Mia", "Evelyn", "Harper", "Camila", "Gianna", "Abigail", "Luna", "Ella", "Elizabeth", "Sofia", "Emily", "Avery", "Mila", "Scarlett", "Eleanor", "Madison", "Layla", "Penelope", "Aria", "Chloe", "Grace", "Ellie", "Nora", "Hazel", "Zoey", "Riley", "Victoria", "Lily", "Aurora", "Violet", "Nova", "Hannah", "Emilia", "Zoe", "Stella", "Everly", "Isla", "Leah", "Lillian", "Addison", "Willow", "Lucy", "Paisley", "Natalie", "Naomi", "Eliana", "Brooklyn", "Elena", "Aubrey", "Claire", "Ivy", "Kinsley", "Audrey", "Maya", "Genesis", "Skylar", "Bella", "Aaliyah", "Madelyn", "Savannah", "Anna", "Delilah", "Serenity", "Caroline", "Kennedy", "Valentina", "Ruby", "Sophie", "Alice", "Gabriella", "Sadie", "Ariana", "Allison", "Hailey", "Autumn", "Nevaeh", "Natalia", "Quinn", "Josephine", "Sarah", "Cora", "Emery", "Samantha", "Piper", "Leilani", "Eva", "Everleigh", "Madeline", "Lydia", "Jade", "Peyton", "Brielle", "Adeline"]),
        LanguageCode.ITALIAN: make_language_data(["Sofia", "Aurora", "Giulia", "Ginevra", "Alice", "Emma", "Giorgia", "Beatrice", "Greta", "Vittoria", "Anna", "Martina", "Chiara", "Matilde", "Ludovica", "Nicole", "Sara", "Bianca", "Camilla", "Arianna", "Noemi", "Gaia", "Francesca", "Mia", "Alessia", "Rebecca", "Elena", "Adele", "Elisa", "Viola", "Marta", "Giada", "Isabel", "Gioia", "Maria", "Emily", "Cecilia", "Azzurra", "Carlotta", "Ambra", "Margherita", "Chloe", "Eleonora", "Asia", "Melissa", "Anita", "Miriam", "Benedetta", "Irene", "Rachele", "Amelia", "Caterina", "Eva", "Diana", "Diletta", "Lucia", "Stella", "Serena", "Alessandra", "Isabella", "Angelica", "Olivia", "Celeste", "Sveva", "Ilaria", "Nina", "Virginia", "Linda", "Agata", "Sophia", "Lavinia", "Anastasia", "Cloe", "Federica", "Zoe", "Clara", "Iris", "Lara", "Gloria", "Nora", "Luna", "Laura", "Clarissa", "Agnese", "Victoria", "Flavia", "Angela", "Valentina", "Claudia", "Letizia", "Rosa", "Elisabetta", "Fatima", "Silvia", "Teresa", "Jasmine", "Maddalena", "Amira", "Lucrezia", "Maya", "Michela", "Livia", "Desiree", "Antonia", "Leonardo", "Francesco", "Lorenzo", "Alessandro", "Andrea", "Mattia", "Gabriele", "Tommaso", "Riccardo", "Edoardo", "Matteo", "Giuseppe", "Nicolo'", "Antonio", "Federico", "Diego", "Davide", "Giovanni", "Pietro", "Samuele", "Christian", "Filippo", "Marco", "Michele", "Luca", "Simone", "Giulio", "Elia", "Alessio", "Salvatore", "Gabriel", "Enea", "Emanuele", "Vincenzo", "Gioele", "Giacomo", "Manuel", "Jacopo", "Daniele", "Thomas", "Samuel", "Cristian", "Giorgio", "Luigi", "Daniel", "Domenico", "Nicola", "Liam", "Nathan", "Raffaele", "Stefano", "Mario", "Angelo", "Ettore", "Noah", "Flavio", "Niccolo'", "Kevin", "Paolo", "Santiago", "Carlo", "Adam", "Alberto", "Alex", "Damiano", "Nicolas", "Nicholas", "Rayan", "Bryan", "Vittorio", "Valerio", "Pasquale", "Amir", "Cesare", "Michael", "Roberto", "Sebastiano", "Enrico", "Sebastian", "Dario", "David", "Achille", "Gregorio", "Brando", "Ludovico", "Massimo", "Fabio", "Antonino", "Ciro", "Dylan", "Carmine", "Martino", "Leo", "Youssef", "Omar", "Cristiano", "Ivan", "Zeno", "Gaetano", "Gennaro", "Mathias", "Ethan", "Claudio"]),
        LanguageCode.DUTCH: make_language_data(["Noah", "Sem", "Liam", "Lucas", "Daan", "Finn", "Levi", "Luuk", "Mees", "James", "Milan", "Sam", "Noud", "Benjamin", "Luca", "Bram", "Mason", "Max", "Thomas", "Adam", "Hugo", "Jesse", "Boaz", "Olivier", "Teun", "Julian", "Lars", "Thijs", "Gijs", "Siem", "Guus", "Mats", "Zayn", "Otis", "Jens", "Jack", "Floris", "Ties", "Vince", "Joep", "David", "Jan", "Stijn", "Sven", "Dex", "Jurre", "Morris", "Ruben", "Owen", "Jayden", "Mohammed", "Tobias", "Moos", "Tim", "Abel", "Tijn", "Jace", "Willem", "Oliver", "Cas", "Fedde", "Ryan", "Jaxx", "Roan", "Quinn", "Xavi", "Daniël", "Dani", "Alexander", "Dean", "Jake", "Sepp", "Ezra", "Mohamed", "Pepijn", "Tom", "Jason", "Aiden", "Jax", "Nathan", "Kai", "Rayan", "Pim", "Oscar", "Elias", "Melle", "Boris", "Mick", "Senn", "Samuel", "Lenn", "Hidde", "Amir", "Johannes", "Job", "Joshua", "Niek", "Tygo", "Stef", "Arthur", "Emma", "Julia", "Mila", "Tess", "Sophie", "Zoë", "Sara", "Nora", "Yara", "Eva", "Liv", "Lotte", "Evi", "Noor", "Anna", "Milou", "Olivia", "Saar", "Lauren", "Nina", "Lieke", "Fleur", "Lynn", "Sofie", "Elin", "Fien", "Nova", "Sarah", "Maud", "Lina", "Mia", "Loïs", "Sofia", "Emily", "Roos", "Fenna", "Ella", "Isa", "Hailey", "Luna", "Hannah", "Julie", "Noa", "Elena", "Sophia", "Bo", "Suze", "Lara", "Maria", "Jasmijn", "Lena", "Esmee", "Cato", "Amy", "Vera", "Lisa", "Liz", "Juul", "Floor", "Hanna", "Norah", "Rosa", "Noé", "Ivy", "Charlotte", "Isabella", "Amber", "Feline", "Elise", "Puck", "Veerle", "Lizzy", "Lize", "Linde", "Livia", "Naomi", "Rosie", "Charlie", "Merel", "Isabel", "Liva", "Fenne", "Anne", "Maeve", "Kiki", "Jill", "Amira", "Benthe", "Iris", "Romy", "Romée", "Eline", "Sanne", "Tessa", "Fiene", "Lola", "Loua", "Femke", "Nola", "Fay"]),
        LanguageCode.FRENCH: make_language_data(["Mathis", "Auguste", "Juliette", "Imran", "Côme", "Léon", "Nour", "Noam", "Samuel", "Chloé", "Lila", "Lana", "Lyam", "Martin", "Clément", "Lisa", "Jeanne", "Eva", "Simon", "Oscar", "Célia", "Malone", "Louna", "Marceau", "Lenny", "Enzo", "Alice", "Elise", "Gabriel", "Zélie", "Salomé", "Axel", "Aaron", "Ambre", "Pablo", "Gabin", "Olivia", "Nino", "Yanis", "Mohamed", "Owen", "Elio", "Apolline", "Gaspard", "Joseph", "Alexandre", "Léa", "Inès", "Maxence", "Margot", "Noah", "Lilou", "Giulia", "Timéo", "Zoé", "Logan", "Emy", "Amélia", "Ella", "Iris", "Maya", "Léna", "Maëlys", "Amir", "Hugo", "Mathilde", "Diego", "Noa", "Clara", "Lyna", "Livio", "Ayden", "Léana", "Noé", "Mila", "Inaya", "Théo", "Héloïse", "Constance", "Gabrielle", "Sofia", "Lou", "Maël", "Mathéo", "Milo", "Roxane", "Kaïs", "Clémence", "Rose", "Manon", "Noémie", "Louis", "Basile", "Lina", "Liam", "Jade", "Jules", "Marius", "Elsa", "Sacha", "Baptiste", "Ali", "Albane", "Ilyan", "Charlie", "Luna", "Léonie", "Soan", "Augustin", "Alya", "Valentine", "Mathys", "Ava", "Louise", "Margaux", "Timothée", "Ayoub", "Soline", "Eliott", "Wassim", "Marie", "Isaac", "Tom", "Evan", "Elena", "Arthur", "Romy", "Anna", "Romane", "Théa", "Anas", "Victoire", "Sarah", "Antoine", "Youssef", "Alba", "Naël", "Nolan", "Lucas", "Léandre", "Assia", "Adèle", "Robin", "Yasmine", "Victoria", "Lise", "Thomas", "Céleste", "Livia", "Adem", "Julia", "Nathan", "Maria", "Nina", "Tiago", "Victor", "William", "Sohan", "Mia", "Ethan", "Alessio", "Capucine", "Lya", "Lyana", "Lily", "Adam", "Thaïs", "Camille", "Aya", "Emmy", "Léo", "Rayan", "Lucie", "Alicia", "Naïm", "Mya", "Lola", "Joy", "Agathe", "Raphaël", "Alix", "Milan", "Emma", "Charlotte", "Paul", "Charly", "Maëlle", "Nora", "Ismaël", "Ilyes", "Malo", "Eden", "Anaïs", "Valentin", "Marin", "Alma", "Maxime", "Ibrahim"]),
        LanguageCode.GERMAN: make_language_data(["Mia", "Emilia", "Hannah", "Emma", "Sophia", "Lina", "Ella", "Mila", "Clara", "Lea", "Marie", "Anna", "Luisa", "Ida", "Leni", "Frieda", "Emily", "Lia", "Lena", "Mathilda", "Charlotte", "Leonie", "Amelie", "Johanna", "Sophie", "Maja", "Lilly", "Lara", "Nele", "Lotta", "Laura", "Sarah", "Nora", "Paula", "Juna", "Mira", "Elisa", "Helena", "Mara", "Melina", "Victoria", "Elena", "Alina", "Maria", "Thea", "Tilda", "Luna", "Marlene", "Carla", "Eva", "Romy", "Pia", "Paulina", "Elina", "Hailey", "Malia", "Merle", "Isabella", "Luise", "Anni", "Antonia", "Josephine", "Zoe", "Fiona", "Pauline", "Julia", "Olivia", "Finja", "Elli", "Lisa", "Martha", "Amalia", "Mina", "Amira", "Carlotta", "Isabell", "Jana", "Rosalie", "Lotte", "Alma", "Liya", "Katharina", "Ronja", "Theresa", "Lucy", "Jasmin", "Aaliyah", "Maila", "Stella", "Elisabeth", "Jule", "Leyla", "Valentina", "Chiara", "Milena", "Annika", "Magdalena", "Amelia", "Melissa", "Amy", "Noah", "Ben", "Matteo", "Finn", "Leon", "Elias", "Paul", "Henry", "Louis", "Felix", "Luca", "Emil", "Jonas", "Theo", "Lukas", "Anton", "Liam", "Maximilian", "Jakob", "Leo", "Oskar", "Max", "Milan", "Karl", "David", "Moritz", "Julian", "Alexander", "Jonah", "Niklas", "Samuel", "Jonathan", "Lio", "Levi", "Mohammed", "Mats", "Raphael", "Aaron", "Philipp", "Tim", "Tom", "Linus", "Erik", "Mika", "Hannes", "Leonard", "Valentin", "Benjamin", "Johann", "Jannis", "Till", "Adrian", "Adam", "Simon", "Milo", "Theodor", "Joshua", "Jan", "Konstantin", "Arthur", "Lian", "Maxim", "Julius", "Vincent", "Fabian", "Marlon", "Nico", "Fiete", "Toni", "Daniel", "Lennard", "Mattis", "Carlo", "Johannes", "Fritz", "Kilian", "Jannik", "Ole", "Malik", "Emilio", "Bruno", "Matti", "Gabriel", "Nils", "Luke", "Lias", "Lasse", "Lenny", "Phil", "Pepe", "Malte", "Levin", "Ludwig", "Nick", "Thilo", "Miran", "Ali", "Benedikt", "Frederik", "Michael"]),
        LanguageCode.SPANISH: make_language_data(["Mario", "Eduardo", "Roberto", "Jaime", "Francisco Jose", "Marcos", "Ignacio", "Hugo", "Alfonso", "Jordi", "Ricardo", "Salvador", "Guillermo", "Gabriel", "Marc", "Emilio", "Mohamed", "Gonzalo", "Julio", "Julian", "Martin", "Jose Miguel", "Tomas", "Agustin", "Nicolas", "Jose Ramon", "Samuel", "Ismael", "Joan", "Cristian", "Felix", "Lucas", "Aitor", "Hector", "Juan Francisco", "Iker", "Alex", "Jose Carlos", "Josep", "Sebastian", "Mariano", "Cesar", "Alfredo", "Domingo", "Jose Angel", "Felipe", "Victor Manuel", "Rodrigo", "Jose Ignacio", "Mateo", "Luis Miguel", "Jose Francisco", "Juan Luis", "Xavier", "Albert", "Maria Carmen", "Maria", "Carmen", "Ana Maria", "Josefa", "Maria Pilar", "Isabel", "Laura", "Maria Dolores", "Maria Teresa", "Ana", "Cristina", "Marta", "Maria Angeles", "Lucia", "Francisca", "Maria Isabel", "Maria Jose", "Antonia", "Dolores", "Sara", "Paula", "Elena", "Maria Luisa", "Raquel", "Rosa Maria", "Pilar", "Manuela", "Concepcion", "Maria Jesus", "Mercedes", "Julia", "Beatriz", "Nuria", "Silvia", "Alba", "Irene", "Rosario", "Juana", "Teresa", "Patricia", "Encarnacion", "Montserrat", "Andrea", "Rocio", "Monica", "Alicia", "Maria Mar", "Rosa", "Sonia", "Sandra", "Marina", "Angela", "Susana", "Natalia", "Yolanda", "Margarita", "Maria Josefa", "Claudia", "Sofia", "Eva", "Carla", "Maria Rosario", "Inmaculada", "Maria Mercedes", "Ana Isabel", "Esther", "Noelia", "Veronica", "Nerea", "Carolina", "Angeles", "Daniela", "Maria Victoria", "Eva Maria", "Ines", "Miriam", "Lorena", "Maria Rosa", "Maria Elena", "Ana Belen", "Victoria", "Maria Concepcion", "Amparo", "Martina", "Maria Antonia", "Alejandra", "Lidia", "Catalina", "Celia", "Maria Nieves", "Consuelo", "Fatima", "Olga", "Ainhoa", "Gloria", "Clara", "Maria Cristina", "Maria Soledad", "Emilia"]),
        LanguageCode.CHINESE: make_language_data(["麗", "偉", "芳", "偉", "秀英", "秀英", "娜", "秀英", "偉", "敏", "静", "麗", "静", "丽", "强", "静", "敏", "敏", "磊", "军", "洋", "勇", "勇", "艳", "杰", "磊", "强", "军", "杰", "娟", "艳", "涛", "涛", "明", "艳", "超", "勇", "娟", "杰", "秀兰", "霞", "敏", "军", "丽", "强", "平", "刚", "杰", "桂英", "芳"]),
    }



import os
import tempfile
from unittest.mock import patch

from intents import Agent
from intents import language

TOY_LANGUAGE_FILE = """
examples:
  - Hi
  - Hello

responses:
  default:
    - text:
      - Greetings, human :)
      - Hi human!
"""

class MockIntentClass:
    name: str = 'test_intent'

    @classmethod
    def parameter_schema(cls):
        return {}

class MockAgentClass(Agent):
    languages = ['en']

def _toy_language_folder(dir_name, intent_name, languages=('en',)):
    for lang in languages:
        lang_dir = os.path.join(dir_name, lang)
        os.makedirs(lang_dir, exist_ok=True)
        with open(os.path.join(lang_dir, intent_name + ".yaml"), 'w') as f:
            print(TOY_LANGUAGE_FILE, file=f)

def test_intent_data_all_languages():
    with tempfile.TemporaryDirectory() as tmp_dir:
        _toy_language_folder(tmp_dir, 'test_intent', ['en'])
        class mock_agent_language_module:
            @staticmethod
            def agent_language_folder(agent_cls):
                return tmp_dir

        with patch('intents.language.intent_language.agent_language', mock_agent_language_module):
            result = language.intent_language_data(MockAgentClass, MockIntentClass)

    assert language.LanguageCode.ENGLISH in result
    assert isinstance(result[language.LanguageCode.ENGLISH], language.IntentLanguageData)

    assert result[language.LanguageCode.ENGLISH].example_utterances == [
        language.ExampleUtterance("Hi", MockIntentClass),
        language.ExampleUtterance("Hello", MockIntentClass)
    ]
    assert result[language.LanguageCode.ENGLISH].slot_filling_prompts == {}
    assert result[language.LanguageCode.ENGLISH].responses == {
        language.IntentResponseGroup.DEFAULT: [
            language.TextIntentResponse(["Greetings, human :)", "Hi human!"])
        ]
    }

# def test_intent_data_skips_private_folders():
#     ...

# def test_intent_data_rich_response_group():
#     ...


"""
It is common for Intents to need some additional business logic to be fulfilled.
For instance, an intent like *"is it going to rain tomorrow?"* should call a
weather service of some sort in order to produce a response.

This is typically achieved through fulfillment calls. Those are functions that
are called by the prediction service after an Intent is predicted, but before
returning the prediction to client, typically through a **REST webhook call**.
Their purpose is to run some logic and optionally change the default Intent
response.

:class:`~intents.model.intent.Intent` classes may override their
:meth:`~intents.model.intent.Intent.fulfill` method to specify the behavior for
fulfillment calls:

.. code-block:: python

    from intents import Intent, FulfillmentContext

    @dataclass
    class UserAsksRain(Intent):
        \"\"\"Is it going to rain tomorrow?\"\"\"
        when: Sys.Date

        def fulfill(self, context: FulfillmentContext, **kwargs) -> Intent:
            result = wheather_api.get_wheather(date=self.when)
            if result == 'rain':
                return AgentReportsRainyWheather()
            else:
                return AgentReportsGoodWeather()

Connectors are responsible for receiving fulfillment requests from Services,
build the appropriate :class:`~intents.model.intent.Intent` instance, call
:meth:`~intents.model.intent.Intent.fulfill` on it, and return its response in
the correct Service format. A **development server** is included, to
conveniently receive REST webhook calls and route them to a Connector; see
:func:`run_dev_server` for details.

We notice that :meth:`~intents.model.intent.Intent.fulfill` returns
:class:`~intents.model.intent.Intent` instances. These are used to trigger a new
intent, that will produce the response. Return `None` if you do not wish to
change the Intent default response.

.. note::

    It is common for platforms to allow fulfillment calls to just produce a set
    of responses as a fulfillment result, without triggering a whole other intent.
    However, for now only triggers are supported, as they are the most general case.
    This may change in next releases.

Fulfillment flow
================
This is the typical flow of a fulfillment request:

#. A prediction request is sent to Service, either from
   :meth:`~intents.connectors.interface.Connector.predict`, or through some native
   service integration (Dialogflow natively supports Telegram, Slack, and so on).
#. Service predicts an Intent and its parameter values
#. Service sends a fulfillment request to its configured endpoint (that's us)
#. *Intents* fulfillment framework receives the reuest and builds a
   :class:`FulfillmentRequest` object.
#. Fulfillment framework passes the :class:`FulfillmentRequest` object to the
   :meth:`~intents.connectors.interface.Connector.fulfill` method of
   :class:`~intents.connectors.interface.Connector`.
#. Connector parses the fulfillment request, and builds both the Intent
   object and a :class:`~intents.model.intent.FulfillmentContext` object
#. Connector calls :meth:`~intents.model.intent.Intent.fulfill` on the intent object, passing the context
#. :meth:`~intents.model.intent.Intent.fulfill` runs its business logic, and
   optionally return another intent to trigger
#. Connector builds the fulfillment response, in a format that Service can understand
#. Connector returns the response to the fulfillment framework
#. Fulfillment framework returns the Connector's response

Serving
=======

The flow above requires you to serve an endpoint that Service can call. For
development you can use the included **development server** (more details at
:func:`run_dev_server`). This is the only function exposed by
:mod:`intents.fulfillment`.

For production you must write your own; this is supposed to be fairly simple:
only thing your endpoint should do is to build a :class:`FulfillmentRequest`
object out of request data, and call
:meth:`~intents.connectors.interface.Connector.fulfill` on a
:class:`~intents.connectors.interface.Connector` instance. The result will be a
dictionary that you can return as it is.

"""
import json
import logging
import http.server

import intents
from intents.helpers.logging import jsondict
from intents.connectors.interface import Connector, FulfillmentRequest

logger = logging.getLogger(__name__)

def run_dev_server(connector: Connector, host: str='', port: str=8000):
    """
    Spawn a simple HTTP server to receive fulfillment requests from the outside.
    This is typically used in combination with some local tunneling solution
    such as `ngrok <https://ngrok.com/>`_

    .. code-block:: python

        from intents.fulfillment import run_dev_server
        from intents.connectors import DialogflowEsConnector, WebhookConfiguration
        from example_agent import ExampleAgent

        webhook = WebhookConfiguration('https://<MY-ADDRESS>.ngrok.io', {"X-Foo": "bar"})
        df = DialogflowEsConnector(..., ExampleAgent, webhook_configuration=webhook)
        df.upload()  # Will set webhook address in Dialogflow
        run_dev_server(df)

    After running the example above, prediction calls (either from
    :meth:`df.predict` or from the Dialogflow UI) will result in Dialogflow
    calling the webhook endpoint for fulfillment. You can try it with the
    intents defined in :mod:`example_agent.calculator`.

    Note that the server will only release its port after its Python process
    dies. This makes it inconvenient to run within a Python CLI, because it
    would be necessary to shut the whole interpreter down at each change. Auto
    reload is not supported yet, your best option is to make a script to
    instantiate Connector and run the server.

    .. warning::

        This server uses Python builtin :mod:`http.server` module, which as per documentation only
        implements basic security check. Also the implementation of request handling is very basic
        and not thoroughly tested. Therefore, it is recommended not to run this service in
        production

    Args:
        connector: A Connector to direct incoming requests to
        host: Optional custom host
        port: Optional custom port
    """
    server_address = (host, port)

    # doesn't work..
    # http.server.HTTPServer.allow_reuse_address = True

    class DevWebhookHttpHandler(http.server.BaseHTTPRequestHandler):

        def do_POST(self):
            content_len = int(self.headers.get('Content-Length'))
            post_body = self.rfile.read(content_len)
            post_body = json.loads(post_body)
            logger.info("POST REQUEST BODY: %s", jsondict(post_body))

            fulfillment_request = FulfillmentRequest(
                body=post_body
            )
            result = connector.fulfill(fulfillment_request)
            logger.info("POST RESPONSE: %s", jsondict(result))
            result = json.dumps(result)
            
            self.send_response(http.server.HTTPStatus.OK)
            self.send_header('Content-type', 'application/json')
            self.end_headers()

            result = bytes(result, 'utf-8')
            self.wfile.write(result)
            self.wfile.flush()

    httpd = http.server.HTTPServer(server_address, DevWebhookHttpHandler)
    print(f"Starting Intents {intents.__version__} development web server on {host}:{port}")
    print("Usage of this module is strongly discouraged in production")
    httpd.serve_forever()


# from example_agent import ExampleAgent
# from intents.connectors import DialogflowEsConnector, WebhookConfiguration
# df = DialogflowEsConnector(
#     '/home/dario/lavoro/dialogflow-agents/_tmp_agents/learning-dialogflow-5827a2d16c34.json',
#     ExampleAgent,
#     default_session='testing-session'
# )
# run_dev_server(df)



"""
Not all Services support the same set of Entities, and therefore it's not
guaranteed that all the entities in :class:`~intents.model.entity.Sys` are
available natively in all the Connectors.

Here we include some built-in custom entities that are used by Connectors for
services that don't provide native support.

.. warning::

    Most of the language content in this module was not reviewed by native
    speakers, nor it was extensively tested. It is recommended to be cautious
    when using them in production.
"""



import logging
import warnings
from dataclasses import dataclass, field

from intents import Intent
from intents.language import IntentResponseDict

logger = logging.getLogger(__name__)

@dataclass
class Prediction:
    """
    One of the core uses of Service Connectors is to predict user utterances, or
    programmatically trigger intents. This class models the result of such
    predictions and triggers.

    You will typically obtain `Prediction` objects from :class:`Connector`
    methods :meth:`~Connector.predict` and :meth:`~Connector.trigger`.

    Args:
        intent: An instance of the predicted Intent
        confidence: A confidence value on the service prediction
        fulfillment_messages: A map of Intent Responses, as they were
            returned by the Service.
        fulfillment_text: A plain-text version of the response
    """
    intent: Intent
    confidence: float
    fulfillment_messages: IntentResponseDict = field(repr=False)
    fulfillment_text: str = None

    @property
    def fulfillment_message_dict(self):
        warnings.warn("Prediction.fulfillment_message_dict is deprecated. Please update "
                      "your code to use Prediction.fulfillment_messages instead.", DeprecationWarning)
        return self.fulfillment_messages



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['intents',
 'intents.connectors',
 'intents.connectors._experimental',
 'intents.connectors._experimental.alexa',
 'intents.connectors._experimental.snips',
 'intents.connectors.dialogflow_es',
 'intents.connectors.interface',
 'intents.helpers',
 'intents.language',
 'intents.model',
 'intents.resources',
 'intents.resources.builtin_entities']

package_data = \
{'': ['*']}

install_requires = \
['dacite>=1.6.0,<2.0.0',
 'google-cloud-dialogflow>=2.0.0,<3.0.0',
 'pyyaml>=5.4.1,<6.0.0']

extras_require = \
{'snips': ['snips-nlu[snips]>=0.20.2,<0.21.0']}

setup_kwargs = {
    'name': 'intents',
    'version': '0.3a1',
    'description': 'Define and operate Dialogflow Agents with a simple, code-first, approach',
    'long_description': '# Intents ⛺\n\n[![Documentation Status](https://readthedocs.org/projects/intents/badge/?version=latest)](https://intents.readthedocs.io/en/latest/?badge=latest)\n[![codecov](https://codecov.io/gh/dariowho/intents/branch/master/graph/badge.svg?token=XAVLW70J8S)](https://codecov.io/gh/dariowho/intents)\n[![HEAD version](https://img.shields.io/badge/head-v0.3a1-blue.svg)](https://img.shields.io/badge/head-v0.3a1-blue.svg)\n[![PyPI version](https://badge.fury.io/py/intents.svg)](https://badge.fury.io/py/intents)\n\n**Intents** is a Python framework to define and operate\nConversational Agents with a simple, code-first approach. *Intents* comes with\nbuilt-in support for Dialogflow ES. Its main benefits are:\n\n* **Agents are Python projects**. You will develop with autocomplete, static type checking\n  and everything you are already used to.\n* **Versioning and CI**. Agents can be versioned on Git, and programmatically\n  deployed just like software.\n* **Human-friendly Connectors**. Intents are classes, predictions are their\n  instances. Support can be extended beyond Dialogflow by implementing custom connectors.\n\nA detailed view of the available features can be found in\n[STATUS.md](STATUS.md). Also, check out the\n[Projects](https://github.com/dariowho/intents/projects) page to keep track of\nrecent developments.\n\n## Install\n\n```sh\npip install intents\n```\n\n## Usage\n\nIntents are defined like standard Python **dataclasses**:\n\n```python\n@dataclass\nclass HelloIntent(Intent):\n    """A little docstring for my Intent class"""\n    user_name: Sys.Person = "Guido"\nMyAgent.register(HelloIntent)\n```\n\nTheir **language** resources are stored in separate YAML files:\n\n```yaml\nutterances:\n  - Hi! My name is $user_name{Guido}\n  - Hello there, I\'m $user_name{Mario}\n\nresponses:\n  default:\n    - text:\n      - Hi $user_name\n      - Hello $user_name, this is Bot!\n```\n\nAgents can be **uploaded** as Dialogflow ES projects directly from code:\n\n```python\ndf = DialogflowEsConnector(\'/path/to/service-account.json\', MyAgent)\ndf.upload()  # You will find it in your Dialogflow Console\n```\n\n*Intents* will act transparently as a **prediction** client:\n\n```python\n>>> predicted = df.predict("Hi there, my name is Mario")\n>>> predicted.intent\nHelloIntent(user_name="Mario")\n>>> predicted.intent.user_name\n"Mario"\n>>> predicted.fulfillment_text\n"Hello Mario, this is Bot!"\n```\n\nFor a complete working example, check out the included [Example Agent](example_agent/). Also, *Intents* **documentation** is published at https://intents.readthedocs.io/ 📚\n\n## Disclaimer\n\n*This project is not affiliated, associated, authorized, endorsed by, or in any way officially connected with Dialogflow. The names Dialogflow, Google, as well as related names, marks, emblems and images are registered trademarks of their respective owners.*\n',
    'author': 'Dario',
    'author_email': 'dario.chi@inventati.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/dariowho/intents',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
Patches :class:`~intents.model.entity.Sys.Language`

Sources:

* https://en.wikipedia.org/wiki/List_of_languages_by_number_of_native_speakers
* https://it.wikipedia.org/wiki/Lingue_per_numero_di_parlanti_madrelingua
* https://de.wikipedia.org/wiki/Liste_der_ISO-639-1-Codes

"""
from typing import List, Union

from intents import Entity, LanguageCode
from intents.language.entity_language import make_language_data

# TODO: only german is complete of all ISO language codes. Spanish and Chinese have fewer
# elements than others

class I_IntentsLanguage(Entity):
    __intents_internal__ = True
    __entity_language_data__ = {
        LanguageCode.ENGLISH: make_language_data([["Chinese", "Mandarin"], "Spanish", "English", "Hindi", "Arabic", "Portuguese", "Bengali", "Russian", "Japanese", "Punjabi", "German", "Javanese", ["Wu", "Shanghainese"], ["Malay", "Indonesian and Malaysian"], "Telugu", "Vietnamese", "Korean", "French", "Marathi", "Tamil", "Urdu", "Turkish", "Italian", ["Yue", "Cantonese"], "Thai", "Gujarati", "Jin", ["Southern Min", "Hokkien and Teochew"], "Persian", "Polish", "Pashto", "Kannada", "Xiang", "Malayalam", "Sundanese", "Hausa", ["Odia", "Oriya"], "Burmese", "Hakka", "Ukrainian", "Bhojpuri", ["Tagalog", "Filipino"], "Yoruba", "Maithili", "Uzbek", "Sindhi", "Amharic", "Fula", "Romanian", "Oromo", "Igbo", "Azerbaijani", "Awadhi", "Gan", ["Cebuano", "Visayan"], "Dutch", "Kurdish", "Serbo-Croatian", "Malagasy", "Saraiki", "Nepali", "Sinhala", "Chittagonian", "Zhuang", "Khmer", "Turkmen", "Assamese", "Madurese", "Somali", "Marwari", "Magahi", "Haryanvi", "Hungarian", "Chhattisgarhi", "Greek", "Chewa", "Deccan", "Akan", "Kazakh", "Northern Min", "Sylheti", "Zulu", "Czech", "Kinyarwanda", "Dhundhari", "Haitian Creole", ["Eastern Min", "Fuzhou dialect"], "Ilocano", "Quechua", "Kirundi", "Swedish", "Hmong", "Shona", "Uyghur", ["Hiligaynon", "Ilonggo", "Visayan"], "Mossi", "Xhosa", "Belarusian", "Balochi", "Konkani"]),
        LanguageCode.ITALIAN: make_language_data([["Cinese", "Mandarino", "cinese mandarino"], "Spagnolo", "Inglese", "Hindi", "Arabo", "Portoghese", ["Bengali", "Bengalese"], "Russo", ["Giapponese", "Nihongo"], ["Cantonese", "Yue"], "Marathi", ["Punjabi", "Punjabi occidentale", "lahnda"], "Telugu", "Turco", ["Coreano", "Hangugo"], "Wu", "Francese", "Tamil", "Tedesco", "Vietnamita", "Urdu", ["Giavanese", "Basa Jawa"], "Lingua araba egiziana", "Italiano", "Gujarati", ["Persiano", "Farsi"], "Bhojpuri", "Yoruba", "Hausa", "Min Nan", "Hakka", ["Indonesiano", "Bahasa Indonesia"], "Kannada", ["Pashto", "lingua pashtun"], "Polacco", "Amarico", ["Xiang", "Hunanese"], ["Malayalam", "Malayāḷaṁ", "Malay", "Malese"], "Oromo", "Maithili", ["Oriya", "Odia"], "Sondanese", "Birmano", "Sindhi", "Curdo", "Ucraino", "Igbo", "Uzbeco", "Rumeno", "Azero", ["Filippino", "Tagalog", "Filipino"], "Olandese", ["Fula", "Fulani"], "Gan", "Thailandese", "Malgascio", ["Serbo-croato", "serbo", "croato", "bosniaco", "montenegrino"], "Magahi", "Rajasthani standard", "Swahili", ["Khmer", "Cambogiano"], "Chhattisgarhi", "Somalo", "Cebuano", "Haryanvi", ["Nepalese", "Nepali"], "Singalese", "Assamese", ["Rangpuri", "Rajbashi"], "Zhuang", "Chittagonian", "Greco", "Kazako", "Deccan", "Ungherese", ["Kinyarwanda", "Orunyarwanda"], "Zulu", "Min Bei", "Shona", "Sylheti", "Ceco", "Uiguro", "Min Dong", ["Chichewa", "Nyanja"], "Kanauji", "Quechua", "Svedese", "Akan", "Bulgaro", ["Xhosa", "isiXhosa"], "Marwari", "Turkmeno", "Creolo haitiano", "Madurese", "Baluchi", "Ilocano", "Varhadi-nagpuri", "Gikuyu", "Konkani", "Armeno", "Danese", "Finlandese", "Ebraico", "Tataro", "Slovacco", "Catalano"]),
        LanguageCode.DUTCH: make_language_data([["Chinees", "Mandarijn"], "Spaans", "Engels", "Hindi", "Arabisch", "Bengaals", "Portugees", "Russisch", "Japans", ["Lahnda", "West-Punjabi"], "Marathi", "Telugu", "Wuyu", "Maleis", "Turks", "Koreaans", "Frans", "Duits", "Vietnamees", "Tamil", "Kantonees", "Urdu", "Javaans", "Italiaans", "Egyptisch-Arabisch", ["Perzisch", "Iraans Perzisch"], "Gujarati", "Bhojpuri", "Minnanyu", "Hakka", "Jinyu", "Hausa", "Kannada", "Indonesisch", "Pools", "Pasjtoe", "Yoruba", "Xiangyu", "Malayalam", "Odia", "Maithili", "Birmaans", "Oost-Punjabi", "Soendanees", "Soedanees Arabisch", "Fula", "Oezbeeks", "Algerijns-Arabisch", "Marokkaans-Arabisch", "Oekraïens", "Igbo", "Noord-Oezbeeks", "Sindhi", "Noord-Levantijns Arabisch", "Roemeens", "Tagalog", "Nederlands", "Azerbeidzjaans", "Saʽidi Arabisch", "Koerdisch", "Ganyu", "Amhaars", "Noord-Pasjtoe", "Magahi", "Thai", "Marwari", "Saraiki"]),
        LanguageCode.FRENCH: make_language_data([["Chinois", "Mandarin", "langues chinoises"], "Espagnol", "Anglais", "Arabe", "Bengali", "Hindi", "Portugais", "Russe", "Japonais", "Français", ["Lahnda", "Pendjabi"], "Javanais", "Allemand", "Vietnamien", "Télougou", "Coréen", "Turc", "Tamoul", "Marathi", "Ourdou", "Italien"]),
        LanguageCode.GERMAN: make_language_data(["Abchasisch", "Afar", "Afrikaans", "Akan", "Albanisch", "Amharisch", "Arabisch", "Aragonesisch", "Armenisch", "Assamesisch", "Avarisch", "Avestisch", "Aymara", "Aserbaidschanisch", "Bambara", "Baschkirisch", "Baskisch", "Weißrussisch", "Bengalisch", "Bihari", "Bislama", "Bosnisch", "Bretonisch", "Bulgarisch", "Birmanisch", ["Katalanisch", "Valencianisch"], "Chamorro", "Tschetschenisch", "Chichewa", "Chinesisch", "Tschuwaschisch", "Kornisch", "Korsisch", "Cree", "Kroatisch", "Tschechisch", "Dänisch", "Dhivehi", ["Niederländisch", "Belgisches Niederländisch"], "Dzongkha", "Englisch", "Esperanto", "Estnisch", "Ewe", "Färöisch", "Fidschi", "Finnisch", "Französisch", "Fulfulde", ["Galicisch", "Galegisch"], "Georgisch", "Deutsch", "Griechisch", "Guaraní", "Gujarati", "Haitianisch", "Hausa", "Hebräisch", "Otjiherero", "Hindi", "Hiri Motu", "Ungarisch", "Interlingua", "Indonesisch", "Interlingue", "Irisch", "Igbo", "Inupiaq", "Ido", "Isländisch", "Italienisch", "Inuktitut", "Japanisch", "Javanisch", ["Grönländisch", "Kalaallisut"], "Kannada", "Kanuri", "Kashmiri", "Kasachisch", "Khmer", "Kikuyu", ["Kinyarwanda", "Ruandisch"], "Kirgisisch", "Komi", "Kikongo", "Koreanisch", "Kurdisch", "oshiKwanyama", "Latein", "Luxemburgisch", "Luganda", ["Limburgisch", "Südniederfränkisch"], "Lingála", "Laotisch", "Litauisch", "Kiluba", "Lettisch", "Manx,", "Manx-Gälisch", "Mazedonisch", ["Malagasy", "Malagassi"], ["Malaiisch", "Malayalam"], "Maltesisch", "Maori", "Marathi", "Marshallesisch", "Mongolisch", "Nauruisch", "Navajo", "Nord-Ndebele", "Nepali", "Ndonga", "Bokmål", "Nynorsk", "Norwegisch", "Yi", "Süd-Ndebele", "Okzitanisch", "Ojibwe", ["Kirchenslawisch", "Altkirchenslawisch"], "Oromo", "Oriya", "Ossetisch", ["Panjabi", "Pandschabi"], "Pali", "Persisch", "Polnisch", "Paschtunisch", "Portugiesisch", "Quechua", ["Bündnerromanisch", "Romanisch"], "Kirundi", "Rumänisch", "Russisch", "Sanskrit", "Sardisch", "Sindhi", "Nordsamisch", "Samoanisch", "Sango", "Serbisch", "Schottisch-gälisch", "Shona", "Singhalesisch", "Slowakisch", "Slowenisch", "Somali", ["Sesotho", "Süd-Sotho"], ["Spanisch", "Kastilisch"], "Sundanesisch", "Swahili", "Siswati", "Schwedisch", "Tamil", "Telugu", "Tadschikisch", "Thai", "Tigrinya", "Tibetisch", "Turkmenisch", "Tagalog", "Setswana", "Tongaisch", "Türkisch", "Xitsonga", "Tatarisch", "Twi", ["Tahitianisch", "Tahitisch"], "Uigurisch", "Ukrainisch", "Urdu", "Usbekisch", "Tshivenda", "Vietnamesisch", "Volapük", "Wallonisch", "Walisisch", "Wolof", "Westfriesisch", "isiXhosa", "Jiddisch", "Yoruba", "Zhuang", "isiZulu"]),
        LanguageCode.SPANISH: make_language_data(["Inglés​", ["Chino", "Chino mandarín​"], "Hindi​", "Español​", "Francés​", "Árabe​", "Bengalí​", "Ruso​", "Portugués​", "Indonesio​", "Urdu​", "Alemán​", "Japonés​", "Suajili​", "Maratí​", "Telugú​", "Turco​", "Chino cantonés​", "Tamil​", "Panyabí occidental​", "Chino Wu​", "Coreano​", "Vietnamita​", "Hausa​", "Javanés​", "Árabe egipcio​", "Italiano​", "Guyarati​", "Tailandés​", "Amárico​", "Canarés​", "Persa iraní​", "Chino Mǐn Nán​"]),
        LanguageCode.CHINESE: make_language_data(["英语", "官話", "印地語", "西班牙语", "法语", "阿拉伯语", "孟加拉语", "俄语", "葡萄牙語", "印尼语", "乌尔都语", "德语", "日语", "斯瓦希里语", "马拉地语", "泰卢固语", "土耳其语", "粤语", "泰米尔语", "旁遮普語", "吴语", "朝鮮語", "越南语", "豪萨语", "爪哇語", "埃及阿拉伯語", "意大利语", "泰语", "古吉拉特语", "康納達語", "波斯语", "博杰普爾語", "閩南語", "菲律賓語"]),
    }



from dataclasses import dataclass
from typing import List
import dacite

@dataclass
class ParseResultIntent:
    intentName: str
    probability: float

@dataclass
class ParseResultSlotRange:
    start: int
    end: int

@dataclass
class ParseResultSlot:
    range: ParseResultSlotRange
    rawValue: str
    value: dict
    entity: str
    slotName: str

@dataclass
class ParseResult:
    input: str
    intent: ParseResultIntent
    slots: List[ParseResultSlot]

def from_dict(parse_result: dict):
    # TODO: handle fallback (intentName=None)
    return dacite.from_dict(ParseResult, parse_result)



import logging
from typing import Union, Dict
from dataclasses import dataclass, field

from intents import Intent
from intents.model.intent import FulfillmentResult

logger = logging.getLogger(__name__)

@dataclass
class FulfillmentRequest:
    """
    The purpose of this class is to uniform fulfillment request payloads, with
    respect to the protocol or framework they are sent with (REST, websocket,
    lambda, ...) 

    Note that the actual parsing comes later, when a :class:`Connector` receives
    the Request, and models it as a :class:`FulfillmentContext`.

    Also, it is not necessary to model a `FulfillmentResponse` counterpart: we
    can assume any fulfillment response can be modeled with a JSON-serializable
    dict.

    Args:
        body: A dict representing the request body
        headers: An optional dict containing the request headers, if present
    """
    body: dict
    headers: dict = field(default_factory=dict)

@dataclass
class WebhookConfiguration:
    """
    Specifies connection parameters for the Agent's webhook. This is the
    endpoint that is called by services to fulfill an intent.

    Args:
        url: An URL to call for fulfillments. Services may require it to be
            HTTPS
        headers: Headers to include in fulfillment requests. Not all services
            may support this
    """
    url: str
    headers: Dict[str, str] = field(default_factory=dict)



import os
import tempfile
from unittest.mock import MagicMock

import pytest

from example_agent import ExampleAgent
from intents.language import LanguageCode, IntentResponseGroup, TextIntentResponse
from intents.helpers import coffee_agent as ca 
from intents.connectors._experimental.snips.connector import SnipsConnector
from intents.connectors._experimental.snips.prediction import SnipsPrediction

def test_export_no_errors():
    c = SnipsConnector(ExampleAgent)
    with tempfile.TemporaryDirectory() as d:
        c.export(d)
        assert os.path.isfile(os.path.join(d, "agent.en.json"))

def test_predict_default_language():
    class MockSnipsEngine:
        def parse(self, message):
            return {
                'input': 'I want an espresso',
                'intent': {'intentName': 'AskEspresso', 'probability': 0.677},
                'slots': []
            }
    c = SnipsConnector(ca.CoffeeAgent)
    c.nlu_engines = {
        LanguageCode.ENGLISH: MockSnipsEngine(),
        LanguageCode.ITALIAN: MockSnipsEngine()
    }
    result = c.predict("Fake text, response is mocked anyway...")
    expected_messages = {
        IntentResponseGroup.DEFAULT: [
            TextIntentResponse(choices=["medium roast espresso, good choice!", "Alright, medium roasted espresso for you"])
        ]
    }
    assert result.intent == ca.AskEspresso()
    assert result.confidence == 0.677
    assert result.fulfillment_messages == expected_messages
    with pytest.warns(DeprecationWarning):
        assert result.fulfillment_message_dict == expected_messages



from datetime import date

from intents.connectors._experimental.alexa import dates

def test_specific_date():
    date_from, date_to = dates.parse_alexa_date("2015-11-25")
    assert date_from == date(2015, 11, 25)
    assert date_to is None

def test_specific_week():
    date_from, date_to = dates.parse_alexa_date("2015-W49")
    assert date_from == date(2015, 11, 30)
    assert date_to == date(2015, 12, 6)

def test_weekend():
    date_from, date_to = dates.parse_alexa_date("2015-W49-WE")
    assert date_from == date(2015, 12, 5)
    assert date_to == date(2015, 12, 6)

def test_month_english():
    date_from, date_to = dates.parse_alexa_date("2018-09")
    assert date_from == date(2018, 9, 1)
    assert date_to == date(2018, 9, 30)

    date_from, date_to = dates.parse_alexa_date("2018-01")
    assert date_from == date(2018, 1, 1)
    assert date_to == date(2018, 1, 31)
    
    date_from, date_to = dates.parse_alexa_date("2008-02")
    assert date_from == date(2008, 2, 1)
    assert date_to == date(2008, 2, 29)

    date_from, date_to = dates.parse_alexa_date("2009-02")
    assert date_from == date(2009, 2, 1)
    assert date_to == date(2009, 2, 28)

def test_month_other_languages():
    date_from, date_to = dates.parse_alexa_date("2018-09-XX")
    assert date_from == date(2018, 9, 1)
    assert date_to == date(2018, 9, 30)

    date_from, date_to = dates.parse_alexa_date("2018-01-XX")
    assert date_from == date(2018, 1, 1)
    assert date_to == date(2018, 1, 31)
    
    date_from, date_to = dates.parse_alexa_date("2008-02-XX")
    assert date_from == date(2008, 2, 1)
    assert date_to == date(2008, 2, 29)

    date_from, date_to = dates.parse_alexa_date("2009-02-XX")
    assert date_from == date(2009, 2, 1)
    assert date_to == date(2009, 2, 28)

def test_year_english():
    date_from, date_to = dates.parse_alexa_date("2008")
    assert date_from == date(2008, 1, 1)
    assert date_to == date(2008, 12, 31)

    date_from, date_to = dates.parse_alexa_date("2009")
    assert date_from == date(2009, 1, 1)
    assert date_to == date(2009, 12, 31)

def test_year_other_languages():
    date_from, date_to = dates.parse_alexa_date("2008-XX-XX")
    assert date_from == date(2008, 1, 1)
    assert date_to == date(2008, 12, 31)

    date_from, date_to = dates.parse_alexa_date("2009-XX-XX")
    assert date_from == date(2009, 1, 1)
    assert date_to == date(2009, 12, 31)

def test_decade():
    date_from, date_to = dates.parse_alexa_date("197X")
    assert date_from == date(1970, 1, 1)
    assert date_to == date(1979, 12, 31)

def test_season():
    date_from, date_to = dates.parse_alexa_date("2021-SP")
    assert date_from == date(2021, 3, 21)
    assert date_to == date(2021, 6, 21)

    date_from, date_to = dates.parse_alexa_date("2021-SU")
    assert date_from == date(2021, 6, 22)
    assert date_to == date(2021, 9, 22)

    date_from, date_to = dates.parse_alexa_date("2021-FA")
    assert date_from == date(2021, 9, 23)
    assert date_to == date(2021, 12, 21)

    date_from, date_to = dates.parse_alexa_date("2021-WI")
    assert date_from == date(2021, 12, 22)
    assert date_to == date(2022, 3, 20)



import logging
from dataclasses import asdict
from collections import ChainMap
from typing import List, Dict, Type

from intents import Intent, EntityMixin
from intents.helpers.data_classes import custom_asdict_factory
from intents.language import intent_language, entity_language
from intents.language import agent_supported_languages, LanguageCode
from intents.connectors._experimental.snips import SnipsConnector
from intents.connectors._experimental.snips import entities as snips_entities
from intents.connectors._experimental.snips import agent_format as af

logger = logging.getLogger(__name__)

def render(connector: SnipsConnector) -> Dict[LanguageCode, dict]:
    languages: List[LanguageCode] = agent_supported_languages(connector.agent_cls)
    result = {}
    for lang in languages:
        try:
            rendered = render_dataset(connector, lang)
            result[lang] = asdict(rendered, dict_factory=custom_asdict_factory())
        except NotImplementedError as exc:
            logger.error("Could not export Agent for language %s: %s", lang, exc)
    return result

def render_dataset(connector: SnipsConnector, lang: LanguageCode) -> af.Dataset:
    return af.Dataset(
        intents=render_all_intents(connector, lang),
        entities=render_all_entities(connector, lang),
        language=lang.value
    )

def render_all_intents(connector: SnipsConnector, lang: LanguageCode):
    """
    Intents without example utterances will break Snips training. Here we filter
    them out.
    """
    result = {}
    for intent in connector.agent_cls.intents:
        rendered = render_intent(connector, intent, lang)
        if rendered:
            result[intent.name] = rendered
    return result

def render_intent(
    connector: SnipsConnector,
    intent_cls: Type[Intent],
    lang: LanguageCode
) -> af.DatasetIntent:
    language_data = intent_language.intent_language_data(connector.agent_cls, intent_cls, lang)
    language_data = language_data[lang]

    if not language_data.example_utterances:
        return None

    # Note that List parameters don't have any special tag. See snips.prediction
    # for details on list parameters
    utterances = []
    for utterance in language_data.example_utterances:
        utterance_data = []
        for chunk in utterance.chunks():
            if isinstance(chunk, intent_language.TextUtteranceChunk):
                utterance_data.append(af.DatasetIntentUtteranceTextSegment(chunk.text))
            elif isinstance(chunk, intent_language.EntityUtteranceChunk):
                chunk: intent_language.EntityUtteranceChunk
                entity_name = snips_entities.ENTITY_MAPPINGS.service_name(chunk.entity_cls)
                utterance_data.append(af.DatasetIntentUtteranceEntitySegment(
                    text=chunk.parameter_value,
                    entity=entity_name,
                    slot_name=chunk.parameter_name
                ))
            else:
                raise ValueError(f"Unsupported utterance chunk type {type(chunk)}. This looks like a bug, please file an issue at https://github.com/dariowho/intents")
        utterances.append(af.DatasetIntentUtterance(
            data=utterance_data
        ))
    return af.DatasetIntent(
        utterances=utterances
    )

def render_all_entities(
    connector: SnipsConnector,
    lang: LanguageCode
) -> Dict[str, af.DatasetEntity]:
    """
    Render all entities in Agent. These are of three types:

    * System entities with builtin support by Snips
    * System entities that are unsupported by Snips
        * Patched with builtin entities by *Intents*
        * Can't be patched, but can be replaced by a placeholder entity
        * Can't be patched or replaced -> NotImplementedError
    * Custom entities
    """
    result = {}
    referenced_sys_entities = connector.agent_cls._referenced_sys_entities

    # Custom Entities
    entities = connector.agent_cls._entities_by_name.values()
    result.update(ChainMap(*[render_entity(connector, e, lang) for e in entities]))

    # Snips-supported Sys entities
    builtin_entities = [e for e in snips_entities.BUILTIN_ENTITIES]
    builtin_entities = [e for e in builtin_entities if e in referenced_sys_entities]
    for e in builtin_entities:
        if not connector.entity_mappings.is_mapped(e, lang):
            raise NotImplementedError(f"Agent references System Entity '{e}', but that is not "
                                      f"supported by Snips NLU for language '{lang}'")
    result.update({snips_entities.ENTITY_MAPPINGS.service_name(e): {} for e in builtin_entities})

    # Patched Sys entities
    patched_mappings = [m for m in snips_entities.PATCHED_MAPPINGS]
    patched_mappings = [m for m in patched_mappings if m.entity_cls in referenced_sys_entities]
    if patched_mappings:
        logger.warning("[%s] The following entities are referenced by Agent, but are not supported natively by Snips: "
                       "%s. They have been patched with Intents builtin entities, which may not provide the "
                       "same quality as native service entities.", lang, [m.entity_cls.name for m in patched_mappings])
        result.update(ChainMap(*[render_entity(connector, m.builtin_entity, lang) for m in patched_mappings]))

    # Placeholder entities
    placeholder_entities = [e for e in snips_entities.PLACEHOLDER_ENTITIES]
    placeholder_entities = [e for e in placeholder_entities if e in referenced_sys_entities]
    if placeholder_entities:
        logger.warning("[%s] The following entities are referenced by Agent, but are not supported natively by Snips: "
                       "%s. They have been replaced with empty placeholder entities, which is likely to lead to poor "
                       "prediction quality.", lang, [e.name for e in placeholder_entities])
    result.update(ChainMap(*[render_placeholder_entity(connector, e) for e in placeholder_entities]))



    return result

def render_entity(
    connector: SnipsConnector,
    entity_cls: Type[EntityMixin],
    lang: LanguageCode
) -> Dict[str, af.DatasetEntity]:
    language_data = entity_language.entity_language_data(connector.agent_cls, entity_cls, lang)
    language_data = language_data[lang]
    entries = []
    for e in language_data:
        entries.append(af.DatasetEntityEntry(value=e.value, synonyms=e.synonyms))
    
    service_name = snips_entities.ENTITY_MAPPINGS.service_name(entity_cls)
    return {
        service_name: af.DatasetEntity(
            data=entries,
            use_synonyms=True,
            automatically_extensible=True,
            matching_strictness=1.0
        )
    }

def render_placeholder_entity(
    connector: SnipsConnector,
    entity_cls: str,
):
    service_name = snips_entities.ENTITY_MAPPINGS.service_name(entity_cls)
    return {
        service_name: af.DatasetEntity(
            use_synonyms=True,
            automatically_extensible=True,
            matching_strictness=1.0,
            data=[]
        )
    }

# from intents.helpers.coffee_agent import CoffeeAgent
# from intents.connectors._experimental.snips import SnipsConnector
# snips = SnipsConnector(CoffeeAgent)
# render(snips)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List

__version__ = "0.3a1"

class SessionEntity:

    _session = None

    def to(self, session: str):
        self._session = session
        return self

from intents.language_codes import LanguageCode

from intents.model.intent import Intent, FulfillmentContext, FulfillmentResult
from intents.model.entity import Entity, EntityMixin, Sys
from intents.model.agent import Agent
from intents.model.relations import follow



import logging
from abc import ABC
from dataclasses import dataclass
from collections import defaultdict
from typing import Dict, List, Union, Tuple

from google.protobuf.json_format import MessageToDict
from google.cloud.dialogflow_v2 import types as pb

from intents.connectors.dialogflow_es import prediction_format as df
from intents.language import IntentResponseGroup, IntentResponseDict, IntentResponse, TextIntentResponse, QuickRepliesIntentResponse, CardIntentResponse

logger = logging.getLogger(__name__)

@dataclass
class DfResponseContextParameter:
    value: Union[str, dict]=None
    original: str=None

@dataclass
class DfResponseContext:
    name: str
    full_name: str
    lifespan: int
    parameters: Dict[str, DfResponseContextParameter]

class PredictionBody:
    """
    This is a superclass for :class:`DetectIntentBody` and
    :class:`WebhookRequestBody`, which both send a `query_result` field
    """
    
    queryResult: df.QueryResult # This field is common in webhook
                                # requests and detectintent responses

    context_lifespans: Dict[str, int]

    def __init__(self, query_result: df.QueryResult):
        self.queryResult = query_result
        self.context_lifespans = {c.simple_name: c.lifespanCount for c in query_result.outputContexts}

    @property
    def intent_name(self):
        return self.queryResult.intent.displayName

    @property
    def intent_parameters(self):
        return self.queryResult.parameters

    def contexts(self) -> Tuple[Dict[str, DfResponseContext], Dict[str, DfResponseContextParameter]]:
        """
        Return a "context_name -> Context" dict of the context in the given
        response, and a dict of global parameter values. In Dialogflow, parameters
        with the same name are overwritten in all contexts, even if they come from
        different intents. Because of this, we can build a map of all global names
        and their values

        TODO: cover less frequent cases (Event/context parameter source,
        event/context/constant parameter default, ...)

        TODO: cache

        :return: A dict of Contexts, A dict of global parameter values
        """
        result_contexts = {}

        for c in self.queryResult.outputContexts:
            parameters: Dict[str, DfResponseContextParameter] = defaultdict(DfResponseContextParameter)
            for p_name, p_value in c.parameters.items():
                if p_name.endswith(".original"):
                    p_name = p_name[:-9]
                    parameters[p_name].original = p_value
                else:
                    parameters[p_name].value = p_value

            context = DfResponseContext(
                name=c.simple_name,
                full_name=c.name,
                lifespan=c.lifespanCount,
                parameters=parameters
            )
            result_contexts[context.name] = context

        result_parameters = _build_global_context_parameters(result_contexts)

        return result_contexts, result_parameters

class DetectIntentBody(PredictionBody):
    """
    This is the result of an intent prediction, exactly as it is sent by
    Dialogflow. Note that the original protobuf is marshaled and converted to
    dataclass, but its field names and content are the same. 
    """
    detect_intent: df.DetectIntentResponse

    def __init__(self, detect_intent_protobuf: pb.DetectIntentResponse):
        detect_intent_dict = MessageToDict(detect_intent_protobuf._pb, including_default_value_fields=True)
        self.detect_intent = df.from_dict(
            data_class=df.DetectIntentResponse,
            data=detect_intent_dict,
        )
        super().__init__(self.detect_intent.queryResult)

class WebhookRequestBody(PredictionBody):
    """
    This is a fulfillment webhook request that Dialogflow sends us.

    Note that, even though there are protobuf schemas for webhook requests
    (`pb.WebhookRequest`), this class is instantiated with a dict. This is
    because webhook requests typically come from REST calls that Dialogflow
    makes to a fulfillment endpoint.
    """
    webhook_request: df.WebhookRequest

    def __init__(self, webhook_request_dict: dict):
        self.webhook_request = df.from_dict(
            data_class=df.WebhookRequest,
            data=webhook_request_dict,
        )
        super().__init__(self.webhook_request.queryResult)

def _build_global_context_parameters(
    contexts: Dict[str, DfResponseContext]
) -> Dict[str, DfResponseContextParameter]:
    result = {}
    for c in contexts.values():
        for p_name, p_value in c.parameters.items():
            if p_name in result and result[p_name] != p_value:
                logger.warning("Parameter '%s' has different value in existing context. " +
                    "This may cause unexpected behavior. Current value: %s. Existing " +
                    "value: %s.", p_name, p_value, result[p_name])
            result[p_name] = p_value
    return result

#
# Response Messages
#

class DialogflowIntentResponse(ABC):
    """
    An interface to add IntentResponse de-serialization from messages in
    Dialogflow Responses. 
    """

    @classmethod
    def from_df_message(cls, df_message: df.QueryResultMessage) -> IntentResponse:
        """
        Build the IntentResponse from a DialogflowResult
        """
        raise NotImplementedError()

class DialogflowTextResponse(TextIntentResponse, DialogflowIntentResponse):

    @classmethod
    def from_df_message(cls, df_message: df.QueryResultMessage) -> TextIntentResponse:
        """
        .. code-block:: json

            {
                "text": {
                    "text": [
                        "Hello, Human"
                    ]
                }
            }
        """
        choices = df_message.text.text
        assert isinstance(choices, list)
        return TextIntentResponse(
            choices=choices
        )

class DialogflowQuickRepliesResponse(QuickRepliesIntentResponse, DialogflowIntentResponse):

    @classmethod
    def from_df_message(cls, df_message: df.QueryResultMessage) -> TextIntentResponse:
        """
        .. code-block: json
            
            {
                "title": "Quick Replies",
                "quickReplies": [
                    "Recommend an hotel",
                    "Send holiday photo"
                ]
            }
        """
        return QuickRepliesIntentResponse(
            replies=df_message.quickReplies.quickReplies
        )

class DialogflowCardResponse(CardIntentResponse, DialogflowIntentResponse):

    @classmethod
    def from_df_message(cls, df_message: df.QueryResultMessage) -> CardIntentResponse:        
        message_buttons = df_message.card.buttons
        card_link = message_buttons[0].postback if message_buttons else None
        return CardIntentResponse(
            title=df_message.card.title,
            subtitle=df_message.card.subtitle,
            image=df_message.card.imageUri,
            link=card_link
        )

def build_response_message(df_message: df.QueryResultMessage):
    """
    Build a fulfillment message from a protobuf structure, as it is found in a
    protobuf response (`query_result.fulfillment_messages`)
    """
    if df_message.text:
        return DialogflowTextResponse.from_df_message(df_message)
    if df_message.quickReplies:
        return DialogflowQuickRepliesResponse.from_df_message(df_message)
    if df_message.card:
        return DialogflowCardResponse.from_df_message(df_message)
    # TODO: custom payload

    raise ValueError(f"Unsupported Fulfillment Message: {df_message}")

def intent_responses(df_body: PredictionBody) -> IntentResponseDict:
    """
    Return the responses in the prediction as standard instances of
    :class:`IntentResponse` and grouped by response group
    (:var:`IntentResponseGroup.DEFAULT` or :var:`IntentResponseGroup.RICH`) 
    """
    result = defaultdict(list)
    for message in df_body.queryResult.fulfillmentMessages:
        if message.platform == df.QueryResultMessagePlatform.PLATFORM_UNSPECIFIED:
            group = IntentResponseGroup.DEFAULT
        else:
            group = IntentResponseGroup.RICH

        result[group].append(build_response_message(message))
    return dict(result)
